// Licensed to the Apache Software Foundation (ASF) under one
// or more contributor license agreements.  See the NOTICE file
// distributed with this work for additional information
// regarding copyright ownership.  The ASF licenses this file
// to you under the Apache License, Version 2.0 (the
// "License"); you may not use this file except in compliance
// with the License.  You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

#include "olap/rowset/unique_rowset_id_generator.h"
#include "util/spinlock.h"
#include "util/uid_util.h"

namespace doris {

UniqueRowsetIdGenerator::UniqueRowsetIdGenerator(const UniqueId& backend_uid) :
    _backend_uid(backend_uid), _inc_id(1) {
}

// generate a unique rowset id and save it in a set to check whether it is valid in the future
RowsetId UniqueRowsetIdGenerator::next_id() {
    std::lock_guard<SpinLock> l(_lock);
    RowsetId rowset_id;
    rowset_id.init(_version, ++_inc_id, _backend_uid.hi, _backend_uid.lo);
    _valid_rowset_ids.insert(rowset_id);
    return rowset_id;
}

bool UniqueRowsetIdGenerator::id_in_use(const RowsetId& rowset_id) {
    std::lock_guard<SpinLock> l(_lock);
    // if rowset_id == 1, then it is an old version rowsetid, not gc it 
    // because old version rowset id is not generated by this code, so that not delete them
    if (rowset_id.version < _version) {
        return true;
    }
    return _valid_rowset_ids.find(rowset_id) != _valid_rowset_ids.end();
}

void UniqueRowsetIdGenerator::release_id(const RowsetId& rowset_id) {
    std::lock_guard<SpinLock> l(_lock);
    _valid_rowset_ids.erase(rowset_id);
}

} // namespace doris
